"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/param/parameter_list.py $
$Id: parameter_list.py 27635 2005-10-28 19:43:18Z dbinger $
"""

from durus.persistent import Persistent
from qp.lib.spec import either, require
from dulcinea.param.parameter import Parameter
from dulcinea.param.parameter_template import ParameterTemplate

class ParameterList (Persistent):
    """
    A special typed list which allows either Parameters or ParameterTemplates.
    """
    list_is = [either(Parameter, ParameterTemplate)]

    def __init__(self, values=None):
        self.list = []
        if values:
            for value in values:
                self.add(value)

    def copy(self):
        result = self.__class__()
        result.list = self.list[:]
        return result

    def __str__(self):
        return "ParameterList(%s)" % self.list

    def __len__(self):
        return len(self.list)

    def __getitem__(self, index):
        return self.list[index]

    def __add__(self, other):
        newlist = ParameterList()
        if isinstance(other, ParameterList):
            newlist.list = self.list + other.list
        else:
            newlist.list = self.list[:]
            for item in other:
                newlist.add(item)
        return newlist

    def get_names(self):
        """() -> [string]

        Return the list of all names in the list.
        """
        return [parameter.get_name() for parameter in self]

    def find(self, name):
        """(name : string) -> any

        Searches the list for a value with name 'name'.  Returns the
        matching value, or None if it couldn't find one.
        """
        for parameter in self.list:
            if name == parameter.get_name():
                return parameter
        require(name, basestring)
        return None

    def add(self, parameter):
        """(parameter : Parameter|ParameterTemplate)

        Raises ValueError if this list already has a value with the same name.
        """
        require(parameter, either(Parameter, ParameterTemplate))
        if self.find(parameter.get_name()) is not None:
            raise ValueError, \
                  "list already has a value named %r" % parameter.get_name()
        self._p_note_change()
        self.list.append(parameter)

    def remove(self, name):
        """(name : string)

        Find the value named by 'name' and remove it.
        """
        require(name, basestring)
        for index, parameter in enumerate(self):
            if name == parameter.get_name():
                self._p_note_change()
                del self.list[index]

    def __iter__(self):
        for item in self.list:
            yield item

    def _get_instances(self, klass):
        return [p for p in self.list if isinstance(p, klass)]

    def get_value_parameters(self):
        """Return a new ParameterList containing all parameters from the
        current list that have values.
        """
        result = ParameterList()
        result.list = self._get_instances(Parameter)
        return result

    def get_templates(self):
        """Return a new ParameterList containing all parameters from the
        current list that don't have values.
        """
        result = ParameterList()
        result.list = self._get_instances(ParameterTemplate)
        return result

    def get_visible(self):
        result = ParameterList()
        result.list = [parameter for parameter in self
                       if not parameter.is_hidden()]
        return result



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/sendmail.py $
$Id: sendmail.py 27873 2006-01-26 17:45:06Z dbinger $

General interface for sending mail to users.
"""
from dulcinea.common import is_qp, get_site_name, get_publisher
from dulcinea.site_util import is_email_enabled, is_live
from qp.mail.rfc822_mailbox import rfc822_mailbox
from quixote.sendmail import sendmail as qx_sendmail

def sendmail(subject, msg_body, to_addrs,
             from_addr=None, reply_to=None, smtp_sender=None,
             cc_addrs=None, extra_headers=None,
             smtp_recipients=None):
    """
    Wrapper for qx_sendmail() function.  Differences:
      * respects the site_util.is_email_enabled(), if it returns
        false then email is not sent
      * always adds "Precedence", "X-No-Archive" headers
      * optionally adds "Reply-To" header
      * if smtp_sender not supplied, defaults to the webmaster
        address from the publisher.
    """
    assert not is_qp()
    if not is_email_enabled():
        return

    # Control over the actual (SMTP) recipient of this message is now in
    # the hands of common_sendmail(), guided by the the site configuration.

    if extra_headers is None:
        extra_headers = []
    if (len(to_addrs) == 1 and
        isinstance(to_addrs[0], basestring) and
        to_addrs[0].startswith("webmaster@")):
        # Looks like this mail was generated by our feedback form.
        # Don't put in the "bulk mail" headers, because they suppress
        # the webmaster autoresponder.
        pass
    else:
        # Add these "bulk mail" headers for all other site-generated mail.
        # They're generally nice to put in auto-generated email.
        extra_headers.extend(["Precedence: junk", "X-No-Archive: yes"])

    if reply_to is not None:
        reply_to = [rfc822_mailbox(args).format() for args in reply_to]
        extra_headers.append("Reply-To: %s" % ", ".join(reply_to))

    if smtp_sender is None:
        smtp_sender = get_publisher().get_webmaster_address()

    if not is_live(get_site_name()):
        assert get_publisher().config.mail_debug_addr is not None
    qx_sendmail(subject, msg_body, to_addrs,
        from_addr=from_addr, extra_headers=extra_headers,
        cc_addrs=cc_addrs, smtp_sender=smtp_sender,
        smtp_recipients=smtp_recipients)






"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/user.py $
$Id: user.py 28604 2006-08-17 19:56:10Z rmasse $
"""
from dulcinea.sort import lexical_sort
from dulcinea.common import get_site_name, get_publisher
from durus.btree import BTree
from durus.persistent import Persistent
from qp.lib.spec import add_getters_and_setters, string, specify
from qp.lib.spec import spec, mapping, pattern, add_getters
from qp.pub.user import User
from qp.lib.util import randbytes
import sha, binascii


def hash_password(password):
    """Apply a one way hash function to a password and return the result."""
    return sha.new(password).hexdigest()

def get_realm():
    return get_site_name()

def get_user_db():
    return get_publisher().get_connection().get_root().get('user_db')

class DulcineaUser (User):
    """
    a registered user.
    """
    global_permissions = {
        "act-as":
            "Allow to act as another user.",
        "create-users":
            "Allow the creation of other users.",
        "manage-permissions":
            "Allow changing of permissions.",
        "staff":
            ("Is a member of the staff, with all of the privileges and "
             "responsibilities thereunto appertaining."),
        "system":
            "Allow to do things normally done by the software system.",
        }

    id_is = spec(
        pattern('^[-A-Za-z0-9_@. ]*$'),
        "unique among users here, spaces allowed")

    def __str__(self):
        return self.id or '*no id*'

    format = __str__ # subclasses should override

    def get_key(self):
        """ used for forming component representing this user in URLs
        """
        return self.get_id()

    def set_id(self, user_id):
        assert self.id == None
        specify(self, id=user_id)

    def set_password(self, new_password, realm=None, check=True):
        """Set the user's password to 'new_password'."""
        if check and self.check_new_password(new_password) != "":
            raise ValueError, 'invalid password'
        digester = self.get_digester()
        if not new_password:
            digester.discard_digest(realm or get_realm())
        else:
            digester.set_digest(realm or get_realm(),
                                hash_password(new_password))

    def has_password(self, password, realm=None):
        """Return true if the provided password is correct."""
        digester = self.get_digester()
        return digester.get_digest(realm or get_realm()) == hash_password(
            password)

    def generate_password (self, realm=None, length=6):
        """Set the password to a random value and return the new password."""
        password = binascii.b2a_base64(binascii.unhexlify(randbytes(length)))
        password = password[:length]
        self.set_password(password, realm=realm)
        return password

    def check_new_password(self, new_password):
        """(string) -> string
        Check if a new password is valid.  Returns the empty string if the
        password is okay otherwise returns a string that describes what is
        wrong with the entered password.
        """
        return ""

    def format_realname(self):
        return ''

    def is_null(self):
        return not self

    def is_disabled(self, realm=None):
        return not bool(self.get_digester().get_digest(realm or get_realm()))

    def is_system(self):
        return self.id == 'SYSTEM'

    def is_admin(self):
        return self.is_granted('staff')

    def is_granted(self, permission, granter=True):
        return self.get_permissions().is_granted(permission, granter)

    def can_manage_permissions(self):
        return self.is_granted('manage-permissions')

    def get_email_realname_pair(self):
        return self.get_email(), self.format_realname() or ''

add_getters_and_setters(DulcineaUser)


class DulcineaUserDatabase (Persistent):
    """
    Class to hold all users in the system.  User IDs are always looked
    up in the user database, so you will generally not be able to use
    a user until it has been added to the user database.
    """

    motd_is = spec(
       string,
       "message-of-the-day")

    user_class = DulcineaUser

    users_is = mapping({string:user_class})

    def __init__(self):
        self.users = BTree()
        self.users[''] = self.create_user('')
        self.motd = ''

    def create_user(self, user_id):
        return self.user_class(user_id)

    def get_all_users(self):
        return self.users.values()

    def __iter__(self):
        return self.users.itervalues()

    def list_users(self, sort=0):
        users = [user for user in self.users.itervalues()
                 if not (user.is_null() or
                         user.is_system() or
                         user.is_disabled())]
        if sort:
            users = lexical_sort(users)
        return users

    def get_disabled_users(self, sort=0):
        users = [user for user in self.users.itervalues()
                 if user.is_disabled() and not (user.is_null() or
                                                user.is_system())]
        if sort:
            users = lexical_sort(users)
        return users

    def get_matching_user(self, identifier):
        """(identifier : string) -> DulcineaUser | None

        Return a user with matching id or email address, or None
        if no such user is found.
        """
        user = self.get_user(identifier)
        if user:
            return user
        elif '@' in identifier:
            identifier = identifier.lower()
            for user in self.users.itervalues():
                if (user.get_email() or '').lower() == identifier:
                    return user
        return None

    def get(self, user_id, default=None):
        return self.users.get(user_id, default=default)

    def get_user(self, user_id):
        """Return the User object with id 'user_id', or None if no such user.
        """
        return self.users.get(user_id)

    def get(self, value, default=None):
        return self.users.get(value, default=default)

    def __getitem__(self, user_id):
        return self.users[user_id]

    def add_user(self, user):
        """Add User object 'user' to the user database.
        """
        assert not self.users.has_key(user.get_id())
        self.users[user.id] = user

    def gen_users_granted(self, permission, granter=True):
        for user in self.users.itervalues():
            if user.is_granted(permission, granter):
                yield user

    def set_motd(self, value):
        specify(self, motd=value)

add_getters(DulcineaUserDatabase)




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/timestamped.py $
$Id: timestamped.py 27308 2005-09-06 17:31:20Z dbinger $
"""
from datetime import datetime
from dulcinea.sort import attr_sort

def timestamp_sorted(timestamped_sequence):
    return attr_sort(timestamped_sequence, 'timestamp')

def reverse_timestamp_sorted(timestamped_sequence):
    result = timestamp_sorted(timestamped_sequence)
    result.reverse()
    return result

class Timestamped(object):

    timestamp_is = datetime

    def __init__(self):
        self.set_timestamp()

    def get_timestamp(self):
        return self.timestamp

    def set_timestamp(self):
        self.timestamp = datetime.now()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/base.py $
$Id: base.py 27812 2006-01-13 16:52:28Z dbinger $

Provides DulcineaPersistent and Copyable.
"""
from durus.persistent import Persistent
from types import ClassType
import new


class Copyable:

    def copy (self):
        klass = self.__class__
        if type(klass) is ClassType:
            newself = new.instance(self.__class__, {})
        elif isinstance(klass, object):
            newself = klass.__new__(klass)
        else:
            raise RuntimeError, "unknown type of class object: %s" % `klass`
        newself.__dict__.update(self.__dict__)
        return newself

    def __copy__ (self):
        return self.copy()


class DulcineaPersistent (Persistent, Copyable):
    pass




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/session.py $
$Id: session.py 28122 2006-04-03 12:32:41Z dbinger $
"""
from dulcinea.attachable import Attachable
from dulcinea.user import DulcineaUser
from qp.pub.session import Session


class DulcineaSession (Session, Attachable):

    effective_user_is = DulcineaUser
    owner_is = DulcineaUser

    def __init__(self):
        Session.__init__(self)
        Attachable.__init__(self)

    def clear_app_state(self):
        """Override to clear any session data
        """

# temporary
from durus.persistent import Persistent
DulcineaSessionManager = Persistent

def update_sessions():
    from datetime import datetime
    from dulcinea.common import get_publisher
    publisher = get_publisher()
    connection = get_publisher().get_connection()
    from qp.lib.spec import get_spec_problems
    sessions = connection.get_root()['session_manager'].sessions
    for session in sessions.values():
        tokens = list(session._form_tokens)[-4:]
        tokenset = session.user.get_tokens()
        tokenset.tokens = (tokenset.tokens + tokens)[-16:]
        tokenset._p_note_change()
        del session._form_tokens
        session.owner = session.last_user
        del session.last_user
        session.effective_user = session.user
        del session.user
        session.remote_address = session._remote_address
        del session._remote_address
        session.access_time = datetime.fromtimestamp(session._access_time)
        del session._access_time
        session.authentication_time = datetime.fromtimestamp(session._creation_time)
        del session._creation_time
        del session.id
        del session.actual_user
        problems = get_spec_problems(session)
        if problems:
            for problem in problems:
                print problem
            assert 0
    from durus.btree import BTree
    new_sessions = BTree()
    connection.get_root()['sessions'] = new_sessions
    for key, value in sessions.items():
        new_sessions[key] = value
    del connection.get_root()['session_manager']
    assert 'session_manager' not in connection.get_root().keys()





"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/persistent_session.py $
$Id: persistent_session.py 28244 2006-04-20 20:52:09Z dbinger $

Provides persistent versions of Quixote's session management classes.
"""
from datetime import datetime
from dulcinea.attachable import Attachable
from dulcinea.common import get_publisher
from dulcinea.user import DulcineaUser
from durus.persistent import Persistent
from durus.persistent_dict import PersistentDict
from durus.persistent_list import PersistentList
from qp.lib.spec import subclass, mapping, sequence, spec, string
from quixote.session import Session as QuixoteSession
from quixote.session import SessionManager as QuixoteSessionManager

def get_session_manager():
    return get_publisher().get_connection().get_root()['session_manager']

class DulcineaSession (Persistent, QuixoteSession, Attachable):
    user_is = spec(
        DulcineaUser,
        "(overrides Session's attribute for stronger typing) "
        "the user object used for access control checks.  Usually "
        "user == actual_user, but when an admin user is acting as some other "
        "user, user is the person being impersonated, and actual_user is "
        "the person actually doing the work (ie.  the admin user)")
    actual_user_is = DulcineaUser
    last_user_is = DulcineaUser
    _form_tokens_is = sequence(string, PersistentList)
    _remote_address_is = (None, string)
    _access_time_is = float
    id_is = string
    _creation_time_is = float

    def __init__ (self, *args):
        if args:
            QuixoteSession.__init__(self, *args)
        Attachable.__init__(self)
        self.clear_authentication()
        self.last_user = self.user
        self._form_tokens = PersistentList()

    def clear_app_state(self):
        """Override to clear any session data when acting as another user
        """
        pass

    def has_info (self):
        return bool(self._form_tokens or self.get_owner())

    def set_actual_user (self, user):
        print 'set_actual_user is deprecated'
        self.set_authenticated(user)

    def set_user(self, user):
        print 'set_user is deprecated'
        self.set_effective_user(user)

    def get_effective_user(self):
        return self.user

    def get_authentication_time(self):
        return datetime.fromtimestamp(self._creation_time)

    def get_access_time(self):
        return datetime.fromtimestamp(self._access_time)

    def set_authenticated(self, user):
        self.user = self.actual_user = user
        if user:
            self.last_user = user

    def set_effective_user(self, user):
        return QuixoteSession.set_user(self, user)

    def get_owner(self):
        return self.last_user

    def clear_authentication(self):
        null_user = get_publisher().get_users().get('')
        self.user = null_user
        self.actual_user = null_user

class DulcineaSessionManager (Persistent, QuixoteSessionManager):

    sessions_is = mapping({string:DulcineaSession}, PersistentDict)
    session_class_is = subclass(DulcineaSession)

    ACCESS_TIME_RESOLUTION = 900 # in seconds (don't dirty the session on
                                 # every hit)

    def __init__ (self, session_class=DulcineaSession, session_mapping=None):
        if session_mapping is None:
            session_mapping = PersistentDict()
        QuixoteSessionManager.__init__(self,
                                       session_class=session_class,
                                       session_mapping=session_mapping)

    def forget_changes (self, session):
        get_publisher().abort()

    def commit_changes (self, session):
        get_publisher().commit()








#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/bin/start-apache.py $
$Id: start-apache.py 27203 2005-08-12 17:45:18Z dbinger $

Script invoked by site command to start apache.
"""

import os
from dulcinea.site_util import get_config_value, get_hostname, \
     get_root_exports, get_pid_file_name, get_administrator_address,\
     parse_address, any_live_sites, any_staging_sites, list_sites, \
     is_live, is_staging, get_base_path, get_docroot
from dulcinea.site_util import get_scgi_off

default_site = list_sites()[0]
sites_directory = get_config_value('sites_directory',
                                   site=default_site,
                                   fallback='')
var_directory = get_config_value('var_directory',
                                 site=default_site,
                                 fallback='')
httpd_conf_name = os.path.join(var_directory, 'generated-httpd.conf')

httpd_conf = open(httpd_conf_name, 'w')

def w (*s):
    httpd_conf.write(' '.join(map(str, s)) + '\n')

def include_site_conf (include_file, site):
    file_name = os.path.join(sites_directory,
                             site,
                             'conf',
                             include_file)
    if os.path.exists(file_name):
        w('Include', file_name)

def include_global_conf (conf_file_name):
    full_name = os.path.join(get_config_value('conf_directory',
                                              site=default_site),
                             'apache',
                             conf_file_name)
    if not os.path.exists(full_name):
        raise SystemExit, 'missing configuration file %r' % full_name
    w('Include', full_name)

def is_apache2():
    return str(get_config_value('apache_version', '1.3',
                                site=default_site)).startswith('2')

def common (site):
    w('ServerName %s' % (get_config_value('servername', site=site) or
                         get_hostname()))
    scgi_address = get_config_value('scgi_address', site=site)
    if scgi_address:
        exports = list(get_root_exports(site))
        if exports:
            exports.sort()
            if '' in exports: # dynamic root
                exports.append('$')
                exports.remove('')
            w('<LocationMatch "^%s/(%s)($|/.*)">' % (get_base_path(site),
                                                     '|'.join(exports)))
            if is_apache2():
                # the apache2 mod_scgi.c requires a slightly different format.
                w('    SCGIServer %s:%s' % parse_address(scgi_address))
            else:
                w('    SCGIServer %s %s' % parse_address(scgi_address))
            w('    SCGIHandler On')
            w('</LocationMatch>')
        else:
            host, port = parse_address(scgi_address)
            base_path = get_base_path(site)
            w('SCGIMount %s %s:%s' % (base_path or '/', host, port))
            for name in get_scgi_off(site):
                w('<LocationMatch "^%s%s">' % (base_path, name))
                w('    SCGIHandler off')
                w('</LocationMatch>')
    include_site_conf('apache-common.conf', site)

w('# Generated. Edit site.conf and/or start_apache.py to change.')
w('ServerAdmin', get_administrator_address(site=default_site)[0])
w('User', get_config_value('daemon_uid', site=default_site))
w('PidFile', get_pid_file_name('apache', site=default_site))

any_live = any_live_sites()
any_staging = any_staging_sites()
assert not (any_live and any_staging)

ports = {}
for site in list_sites():
    for option in ('http_address', 'https_address'):
        option_value = get_config_value(option, site=site)
        if option_value:
            ip, port = parse_address(option_value)
            ports[port]=1
for port in ports:
    w('Listen', port)


if not any_live:
    # Directives if not live for any site.
    w('CoreDumpDirectory', get_config_value('var_directory', site=default_site))
    w('LogLevel debug')

include_global_conf('apache-global.conf')

print
for site in list_sites():
    if not get_config_value('http_address', site=site):
        continue
    if (any_live and not is_live(site)):
        # In live mode, skip sections that aren't live 
        continue
    if (any_staging and not is_staging(site)):
        # In staging mode, skip sections that aren't staging
        continue

    docroot = get_docroot(site)
    if not os.path.exists(docroot):
        raise SystemExit, 'docroot directory %r does not exist' % docroot
    host_iface = get_config_value('http_address', site=site)
    host_ssl_iface = get_config_value('https_address', site=site)

    w('\n<VirtualHost %s>' % host_iface, '# %s http' % site)
    print ' ', site, host_iface,
    w('DocumentRoot ' + docroot)
    common(site)
    include_site_conf('apache-http.conf', site)
    w('</VirtualHost>')
    if host_ssl_iface:
        print host_ssl_iface,
        w('\n<VirtualHost %s>' % host_ssl_iface,
          '# %s https' % site)
        w('DocumentRoot ' + docroot)
        common(site)
        w('SSLEngine on')
        w('SSLOptions +StdEnvVars')
        ssl_cert = get_config_value('ssl_certificate', site=site)
        ssl_cert_key = ssl_cert.replace('.crt', '.key')
        w('SSLCertificateFile %s' % ssl_cert)
        w('SSLCertificateKeyFile %s' % ssl_cert_key)
        w('</VirtualHost>')
    if is_live(site):
        print 'live',
        include_site_conf('apache-live.conf', site)
    print

httpd_conf.close()

httpd = get_config_value('httpd', site=default_site)
print "  " + httpd + " -f " + httpd_conf_name
os.execv(httpd, [httpd, '-f', httpd_conf_name])




#!/www/python/bin/python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/bin/start-scgi.py $
$Id: start-scgi.py 27929 2006-02-01 18:43:16Z rmasse $

Script invoked by site command to start scgi.
"""
import os, sys
site = os.environ['SITE'] = sys.argv[1]
from dulcinea.common import create_publisher as common_create_publisher
from dulcinea.site_util import enable_email, log_to, get_base_path
from dulcinea.site_util import ensure_uid_gid_not_root, parse_address, is_devel
from dulcinea.site_util import get_administrator_address, is_live, is_staging
from dulcinea.site_util import get_config_value, get_pid_file_name, is_local
from durus.logger import logger
from quixote.config import Config
from quixote.server.scgi_server import run as run_scgi
import pwd
import socket

ensure_uid_gid_not_root()
pid_file_name = get_pid_file_name('scgi', site)
ip, port = parse_address(
    get_config_value('scgi_address', site=site))
site_log_dir = os.path.join(get_config_value('log_directory', site=site), site)
error_log = os.path.join(site_log_dir, 'error.log')
assert is_local(get_config_value('scgi_address', site=site))
logger.setLevel(int(get_config_value('logginglevel', fallback=20, site=site)))

def create_publisher():
    # set config options
    config = Config()
    config.form_tokens = 1
    config.display_exceptions = 0
    config.session_cookie_path = '/'
    site_log_dir = os.path.join(get_config_value('log_directory', site=site),
                                site)
    config.access_log = os.path.join(site_log_dir, 'access.log')
    config.error_log = error_log
    try:
        import zlib
        zlib # to quiet import checker
        config.compress_pages = 1
    except ImportError:
        config.compress_pages = 0
    administrator = get_administrator_address()

    if is_live(site):
        config.error_email = administrator[0]
        config.mail_from = administrator
    elif is_staging(site):
        config.error_email = administrator[0]
        config.mail_from = administrator
        config.mail_debug_addr = administrator[0]
    else: # devel
        config.display_exceptions = 1
        user = pwd.getpwuid(os.geteuid())[0] # Owner of the SCGI process
        user_email = user + '@' + socket.getfqdn()
        config.mail_debug_addr = get_config_value(
            'mail_debug_addr',
            fallback=user_email,
            site=site)
        config.mail_from = user_email

    # Adjust any values that are set in the site configuration.
    for config_var in config.config_vars:
        if get_config_value(config_var, site=site):
            setattr(config, config_var,
                    get_config_value(config_var, site=site))
    if get_config_value('email_enabled', site=site) is True:
        enable_email()
    publisher = common_create_publisher(config=config)
    if not is_live(site):
        assert publisher.config.mail_debug_addr is not None
    from dulcinea.site_util import get_cache_size
    publisher.get_connection().set_cache_size(get_cache_size(site))
    publisher.log("SCGI server for %s started (pid %d)" % (site, os.getpid()))
    publisher.log("Durus object cache size is %s" %
                  publisher.get_connection().get_cache_size())
    return publisher


log_to(error_log)

if is_devel(site):
    max_children = 1 # faster for development
else:
    max_children = 5

max_children = get_config_value('max_children', site=site,
                                fallback=max_children)

pid = os.fork()
if pid == 0:
    pid = os.getpid()
    pidfile = open(pid_file_name, 'w')
    pidfile.write(str(pid))
    pidfile.close()
    try:
        run_scgi(create_publisher,
                 port=port,
                 max_children=max_children,
                 script_name=get_base_path())
    finally:
        # grandchildren get here too, don't let them unlink the pid
        if pid == os.getpid():
            try:
                os.unlink(pid_file_name)
            except OSError:
                pass



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_user_database.py $
$Id: utest_user_database.py 28600 2006-08-17 14:13:32Z rmasse $
"""
from dulcinea.contact import ContactDatabase, Contact
from qp.lib.spec import get_spec_problems
from dulcinea.user import DulcineaUser, DulcineaUserDatabase
from sancho.utest import UTest, raises
from dulcinea.common import get_site_name

try:
    get_site_name()
except AssertionError:
    import os
    os.environ['SITE'] = 'test'
    get_site_name()

class Test (UTest):

    def check_user_database(self):
        user1 = DulcineaUser('1')
        user1.set_password('foo')
        user2 = DulcineaUser('2')
        self.db = DulcineaUserDatabase()
        self.db.add_user(user1)
        assert self.db.list_users() == [user1]
        for user in self.db:
            assert user is user1 or user.is_null()
        self.db.add_user(user2)
        assert self.db.get_disabled_users() == [user2]
        assert self.db.list_users() == [user1]
        assert self.db.get('1') is user1
        assert self.db.get('3', default='g') == 'g'
        user2.set_password('junk')
        assert self.db.list_users(sort=True) == [user1, user2]
        assert self.db.get_user('1') == user1
        assert self.db['1'] == user1
        raises(AssertionError, self.db.add_user, user1)
        assert self.db.get_motd() == ''
        self.db.set_motd('The message')
        assert self.db.get_motd() == 'The message'

    def check_contact_database(self):
        self.db = ContactDatabase()
        assert get_spec_problems(self.db) == [], get_spec_problems(self.db)
        user1 = Contact('1')
        user1.set_email('foo@bar.com')
        self.db.add_user(user1)
        assert self.db.get_matching_user('notfound@bar.com') == None
        assert self.db.get_matching_user('foo@bar.com') == user1

if __name__ == "__main__":
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/util.py $
$Id: util.py 28320 2006-05-02 19:46:16Z rmasse $
"""
from datetime import datetime
from distutils.fancy_getopt import wrap_text
from dulcinea.common import htmlescape, htmltext, stringify
from formatter import AbstractFormatter, DumbWriter
from htmllib import HTMLParser
from os.path import isdir, dirname, abspath
from StringIO import StringIO
import re
import time


SAFE_TAGS = ['<p>', '</p>', '<b>', '</b>', '<i>', '</i>', '<ul>',
             '</ul>', '<li>', '</li>', '<br>', '<pre>', '</pre>',
             '<strong>', '</strong>']

def _htmlescape_except(text):
    """(text) -> htmltext

    Quotes 'text' for use in an HTML page, except for the tags
    listed in 'safe_tags' which are considered safe (e.g. <p>, <b>).
    Both upper and lower case versions of the tags will be applied
    """
    if text is None:
        return None
    text = stringify(htmlescape(text))
    for tag in SAFE_TAGS:
        tag_re = re.compile(stringify(htmlescape(tag)), re.IGNORECASE)
        text = tag_re.sub(tag, text)
    return htmltext(text)

# re to automatically hyperlink email addresses and URLs
_link_re = re.compile(r"""\b(
    # email address
      [\w.-]+ # local part
      @
      [\w.-]+\.[\w]{2,4} # domain
    |
    # URL
      (?:https?://|www\.) # must start with http or www
      [\w.-]+\.[\w]{2,4} # domain
      (?::\d+)? # optional port
      (?:/[\w#$%&+,-./:;=?@\[\]^_|~]*)? # optional path
   )""", re.VERBOSE)

def activate_links(text):
    """(text:str|None) -> htmltext|None
    Returns an htmlescaped version of text, with things that look like email
    addresses and URLs turned into hyperlinks.
    """
    def _link_replace(m):
        text = url = m.group(0)
        extra = ""
        if text.find("@") == -1:
            if text[-1] in ".,":
                # don't include as part of the URL (easier to handle here
                # than in the regex)
                extra = text[-1]
                text = text[:-1]
            if not text.startswith("http"):
                url = "http://" + text
        else:
            if not text.startswith("mailto"):
                url = "mailto:" + text
        return '<a href="%s">%s</a>%s' % (url, text, extra)

    if text is None:
        return None
    return htmltext(_link_re.sub(_link_replace, stringify(htmlescape(text))))


def format_text(text, safe_tags=None):
    """(text : string, safe_tags : [str]) -> htmltext

    Convert an ASCII string containing text into HTML.
    The resulting HTML has been reformatted in the following ways:
       * <, &, and > are escaped
       * Things that look like email addresses and URLs will be turned into
         hyperlinks.
       * If safe_tags is defined, it is expected to be a list of tags
         considered to be safe (e.g. <p>, <b>, etc).  These tags will
         not be escaped.
    """
    return activate_links(_htmlescape_except(text))

def sanitize_url(url):
    """(url:string) -> string
    Try to ensure a URL is well-formed, by adding http:// if it isn't present.
    """
    if url is None:
        return None
    if '@' in url and not url.startswith('mailto:'):
        # assume it's an e-mail address
        url = "mailto:" + url
    elif url.find(":") == -1:
        # assume http:// is missing
        url = "http://" + url
    return url

_paragraph_re = re.compile('\n\n+')

def split_paragraphs(text):
    return _paragraph_re.split(text)

def wrap_paragraphs(text):
    """(text) -> string
    Wrap a sequence of paragraphs for output as plain text.
    """
    if text is None:
        return ""
    return '\n\n'.join([wrap_paragraph(paragraph)
                        for paragraph in split_paragraphs(text)])

def wrap_paragraph(text):
    line_length = 70
    if isinstance(text, unicode):
        lines = [text.__class__(x, 'utf-8')
                 for x in wrap_text(text.encode('utf-8'), line_length)]
    else:
        lines = wrap_text(text, line_length)
    return '\n'.join(lines)

def insert_paragraph_tags(text):
    """(text:string|None) -> htmltext|None

    Prepare a text field for display as HTML.
    Currently this just HTML quotes the string and then inserts
    <p> tags at blank lines.  
    """
    if text is None:
        return None
    return htmltext(
        '<p>' +
        _paragraph_re.sub('</p><p>', stringify(htmlescape(text))) +
        '</p>')

def datetime_to_int(date_time):
    """(date_time:datetime) -> int
    Returns the number of seconds since the epoch.
    """
    return int(time.mktime(date_time.timetuple()))

def beginning_of_next_month(date_time):
    """(date_time:datetime) -> datetime
    Return a datetime for the exact beginning of the month
    following the given date_time.  
    """
    year = date_time.year
    month = date_time.month + 1
    if month == 13:
        year += 1
        month = 1
    return datetime(year=year, month=month, day=1, hour=0, minute=0,
                    second=0, microsecond=0)

def html2txt(text):
    """(any) -> str
    """
    sio = StringIO()
    parser = HTMLParser(AbstractFormatter(DumbWriter(sio)))
    parser.feed(stringify(text))
    return sio.getvalue()

def is_new(persistent_object):
    """(persistent_object : durus.persistent.Persistent) -> boolean
    """
    return persistent_object._p_connection is None

def get_module_directory(module):
    result = abspath(module.__file__)
    if isdir(result):
        return result
    else:
        return dirname(result)

import re
pat = re.compile(r'([\x80-\xff]{2,}(?:"|\(tm\))?)')
utf8indicator = re.compile(r'[\xc2-\xc3]')
cp1252indicator = re.compile(r'[\x80-\x9f]')

def get_targets(s):
    return [w for w in pat.findall(s)
            if utf8indicator.search(w)]

def ureduce(s):
    for depth in range(10):
        try:
            s = unicode(s, 'utf8')
            try:
                s = s.encode('latin1')
            except UnicodeEncodeError:
                return s
        except UnicodeDecodeError:
            pass
    return s

def demangle(s):
    if s.endswith('(tm)'):
        s = s.replace('(tm)', '\xc3\x82\xc2\x99')
    if s.endswith('"'):
        s = '"'
    if s.endswith('-'):
        s = '-'
    return s

def find_using_utf8(s):
    orig = s
    if not utf8indicator.search(s):
        return s
    for target in get_targets(s):
        result = ureduce(demangle(target))
        if isinstance(result, unicode):
            result = result.encode('utf8')
        s = s.replace(target, result)
    try:
        return unicode(s, 'utf8')
    except UnicodeDecodeError:
        return orig

def unify_instance(obj):
    from qp.lib.spec import get_specs, match
    for name, spec in get_specs(obj.__class__).items():
        if match(u'', spec):
            old_value = getattr(obj, name)
            if not isinstance(old_value, str):
                continue
            # Try ascii first.
            try:
                setattr(obj, name, unicode(old_value, 'ascii'))
                continue
            except UnicodeDecodeError:
                pass
            # try hyper utf8.
            result = find_using_utf8(old_value)
            if isinstance(result, unicode):
                setattr(obj, name, result)
                if u'\xc3' in result or u'\xc2' in result:
                    print 'hyper utf8', obj
                continue
            if cp1252indicator.search(old_value):
                try:
                    setattr(obj, name, unicode(old_value, 'cp1252'))
                    continue
                except UnicodeDecodeError:
                    pass
            if utf8indicator.search(old_value):
                try:
                    setattr(obj, name, unicode(old_value, 'utf16'))
                    continue
                except UnicodeDecodeError:
                    pass
                print 'utf8', obj
            setattr(obj, name, unicode(old_value, 'latin1'))

def unify(connection):
    from durus.connection import gen_every_instance
    from durus.persistent import Persistent
    for item in gen_every_instance(connection, Persistent):
        unify_instance(item)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/param/test/utest_types.py $
$Id: utest_types.py 28105 2006-03-30 16:24:19Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.material import get_material_db
from dulcinea.param import get_parameter_db
from dulcinea.param.errors import ParamTypeError
from dulcinea.param.parameter_type import ParameterType
from dulcinea.physical_unit import get_standard_unit
from dulcinea.physical_value import PhysicalValue
from dulcinea.range_value import RangeValue
from sancho.utest import raises

class ParameterTypeTest (CommonTest):

    def _pre (self):
        CommonTest._pre(self)
        if get_parameter_db() is None:
            return
        self.plib = get_parameter_db()
        self.mdb = get_material_db()

    def check_int_type (self):
        if get_parameter_db() is None:
            return
        parameter_type1 = self.plib.get_parameter_type(ParameterType.ATOMIC,
                                                       ParameterType.INT)

        assert parameter_type1.is_atomic()
        assert parameter_type1.is_int()
        assert not parameter_type1.is_string()
        assert not parameter_type1.is_physical_value()
        assert not parameter_type1.is_material()
        assert not parameter_type1.is_list()
        assert not parameter_type1.is_table()
        assert not parameter_type1.is_boolean()

        parameter_type2 = self.plib.get_parameter_type(ParameterType.ATOMIC,
                                                       ParameterType.STRING)
        assert parameter_type2 is not parameter_type1
        parameter_type2 = self.plib.get_parameter_type(ParameterType.ATOMIC,
                                                       ParameterType.INT)
        assert parameter_type2 is parameter_type1

        parameter_type1.check_value(None)
        parameter_type1.check_value(10)
        parameter_type1.check_value(RangeValue(10, 20))
        raises(ParamTypeError, parameter_type1.check_value, [10, 20])
        raises(ParamTypeError, parameter_type1.check_value, 'hello')
        raises(ParamTypeError, parameter_type1.check_value, ['hello', 'bye'])
        raises(AssertionError,
            self.plib.get_parameter_type,
            ParameterType.ATOMIC, ParameterType.INT, key_type=1)

    def check_string_type (self):
        if get_parameter_db() is None:
            return
        parameter_type1 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.STRING)

        assert parameter_type1.is_atomic()
        assert not parameter_type1.is_int()
        assert parameter_type1.is_string()
        assert not parameter_type1.is_physical_value()
        assert not parameter_type1.is_material()
        assert not parameter_type1.is_list()
        assert not parameter_type1.is_table()
        assert not parameter_type1.is_boolean()

        parameter_type2 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.INT)
        assert parameter_type2 is not parameter_type1
        parameter_type2 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.STRING)
        assert parameter_type2 is parameter_type1

        parameter_type1.check_value(None)
        parameter_type1.check_value('hello')
        raises(ParamTypeError, parameter_type1.check_value, ['hello', 'bye'])
        raises(ParamTypeError, parameter_type1.check_value, 10)
        raises(ParamTypeError, parameter_type1.check_value, RangeValue(10, 20))
        raises(ParamTypeError, parameter_type1.check_value, [10, 20])

    def check_boolean_type (self):
        if get_parameter_db() is None:
            return
        parameter_type1 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.BOOLEAN)

        assert parameter_type1.is_atomic()
        assert not parameter_type1.is_int()
        assert not parameter_type1.is_string()
        assert not parameter_type1.is_physical_value()
        assert not parameter_type1.is_material()
        assert not parameter_type1.is_list()
        assert not parameter_type1.is_table()
        assert parameter_type1.is_boolean()

        parameter_type2 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.INT)
        assert parameter_type2 is not parameter_type1
        parameter_type2 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.BOOLEAN)
        assert parameter_type2 is parameter_type1

        parameter_type1.check_value(True)
        parameter_type1.check_value(False)
        parameter_type1.check_value(0)
        parameter_type1.check_value(1)
        parameter_type1.check_value(None)
        raises(ParamTypeError, parameter_type1.check_value, '')
        raises(ParamTypeError, parameter_type1.check_value, [True])
        assert parameter_type1.format_value(True) == 'yes'
        assert parameter_type1.format_value(0) == 'no'

    def check_physical_value_type (self):
        if get_parameter_db() is None:
            return
        parameter_type1 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.PHYSICAL_VALUE)

        assert parameter_type1.is_atomic()
        assert not parameter_type1.is_int()
        assert not parameter_type1.is_string()
        assert parameter_type1.is_physical_value()
        assert not parameter_type1.is_material()
        assert not parameter_type1.is_list()
        assert not parameter_type1.is_table()
        assert not parameter_type1.is_boolean()

        parameter_type2 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.PHYSICAL_VALUE)
        assert parameter_type2 is parameter_type1

        inch = get_standard_unit('inch')
        parameter_type2 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.PHYSICAL_VALUE,
            element_sub_type=inch)
        assert parameter_type2 is not parameter_type1

        parameter_type1.check_value(None)
        parameter_type1.check_value(PhysicalValue(10))
        parameter_type1.check_value(PhysicalValue(RangeValue(10, 20)))
        parameter_type2.check_value(PhysicalValue(10, 'inch'))
        parameter_type2.check_value(PhysicalValue(RangeValue(10, 20), 'inch'))
        parameter_type2.check_value(PhysicalValue(10, 'm'))
        parameter_type2.check_value(PhysicalValue(RangeValue(10, 20), 'm'))
        raises(ParamTypeError, parameter_type1.check_value, 
            PhysicalValue(10, 'inch'))
        raises(ParamTypeError, parameter_type2.check_value, PhysicalValue(10))
        raises(ParamTypeError, parameter_type2.check_value, 
            PhysicalValue(10, 's'))
        raises(ParamTypeError, parameter_type1.check_value, [PhysicalValue(10)])
        raises(ParamTypeError, parameter_type2.check_value,
            [PhysicalValue(10, 'inch')])
        raises(ParamTypeError,
            parameter_type1.check_value, 10)
        raises(ParamTypeError,
            parameter_type1.check_value, RangeValue(10, 20))
        raises(ParamTypeError,
            parameter_type1.check_value, [10, 20])
        raises(ParamTypeError,
            parameter_type1.check_value, 'hello')
        raises(ParamTypeError,
            parameter_type1.check_value, ['hello', 'bye'])
        raises(AssertionError,
            self.plib.get_parameter_type,
            ParameterType.ATOMIC, ParameterType.PHYSICAL_VALUE,
            element_sub_type=1)
        raises(AssertionError,
            self.plib.get_parameter_type,
            ParameterType.ATOMIC, ParameterType.PHYSICAL_VALUE, key_type=1)
        raises(AssertionError,
            self.plib.get_parameter_type,
            ParameterType.ATOMIC, ParameterType.PHYSICAL_VALUE, key_sub_type=1)

    def check_material_type (self):
        if get_parameter_db() is None:
            return
        parameter_type1 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.MATERIAL)

        assert parameter_type1.is_atomic()
        assert not parameter_type1.is_int()
        assert not parameter_type1.is_string()
        assert not parameter_type1.is_physical_value()
        assert parameter_type1.is_material()
        assert not parameter_type1.is_list()
        assert not parameter_type1.is_table()
        assert not parameter_type1.is_boolean()

        parameter_type2 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.MATERIAL)
        assert parameter_type2 is parameter_type1

        silicone = self.mdb.get_material('silicone')
        gold = self.mdb.get_material('gold')
        inch = get_standard_unit('inch')

        parameter_type2 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.MATERIAL)
        assert parameter_type2 is parameter_type1

        parameter_type1.check_value(None)
        parameter_type1.check_value(silicone)
        parameter_type1.check_value(gold)
        raises(ParamTypeError,
            parameter_type1.check_value, [gold])
        raises(ParamTypeError,
            parameter_type1.check_value, 'hello')
        raises(ParamTypeError,
            parameter_type1.check_value, ['hello', 'bye'])
        parameter_type2.check_value(silicone)


    def check_list_type (self):
        if get_parameter_db() is None:
            return
        parameter_type1 = self.plib.get_parameter_type(
            ParameterType.LIST, ParameterType.INT)
        assert not parameter_type1.is_atomic()
        assert parameter_type1.is_int()
        assert not parameter_type1.is_string()
        assert not parameter_type1.is_physical_value()
        assert not parameter_type1.is_material()
        assert parameter_type1.is_list()
        assert not parameter_type1.is_table()
        assert not parameter_type1.is_boolean()

        parameter_type2 = self.plib.get_parameter_type(
            ParameterType.LIST, ParameterType.INT)
        assert parameter_type2 is parameter_type1
        parameter_type2 = self.plib.get_parameter_type(
            ParameterType.LIST, ParameterType.STRING)
        assert parameter_type2 is not parameter_type1
        parameter_type3 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.STRING)
        assert parameter_type3 is parameter_type2.element_type

        assert not parameter_type2.is_int()
        assert parameter_type2.is_string()

        parameter_type1.check_value(None)
        parameter_type1.check_value([10])
        parameter_type1.check_value([10, 20])
        parameter_type1.check_value([RangeValue(10, 20)])
        parameter_type1.check_value([10, 20, RangeValue(10, 20)])
        raises(ParamTypeError,
            parameter_type1.check_value, 10)
        raises(ParamTypeError,
            parameter_type1.check_value, RangeValue(10, 20))
        raises(ParamTypeError,
            parameter_type1.check_value, 'hello')

    def check_table_type (self):
        if get_parameter_db() is None:
            return
        parameter_type1 = self.plib.get_parameter_type(
            ParameterType.TABLE, ParameterType.PHYSICAL_VALUE,
            key_type=ParameterType.MATERIAL)

        assert not parameter_type1.is_atomic()
        assert not parameter_type1.is_int()
        assert not parameter_type1.is_string()
        assert parameter_type1.is_physical_value()
        assert not parameter_type1.is_material()
        assert not parameter_type1.is_list()
        assert parameter_type1.is_table()
        assert not parameter_type1.is_boolean()

        assert not parameter_type1.is_key_int()
        assert not parameter_type1.is_key_string()
        assert not parameter_type1.is_key_physical_value()
        assert parameter_type1.is_key_material()

        silicone = self.mdb.get_material('silicone')
        gold = self.mdb.get_material('gold')
        inch = get_standard_unit('inch')

        parameter_type2 = self.plib.get_parameter_type(
            ParameterType.TABLE, ParameterType.PHYSICAL_VALUE,
            key_type=ParameterType.MATERIAL)
        assert parameter_type2 is parameter_type1
        parameter_type2 = self.plib.get_parameter_type(
            ParameterType.TABLE, ParameterType.PHYSICAL_VALUE,
            key_type=ParameterType.MATERIAL)
        assert parameter_type2 is parameter_type1
        parameter_type3 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.PHYSICAL_VALUE)
        assert parameter_type3 is parameter_type1.element_type
        assert parameter_type3 is parameter_type2.element_type
        parameter_type3 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.MATERIAL)
        assert parameter_type3 is parameter_type1.key_type
        assert parameter_type3 is parameter_type2.key_type
        parameter_type3 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.MATERIAL)
        assert parameter_type3 is parameter_type1.key_type
        assert parameter_type3 is parameter_type2.key_type
        parameter_type3 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.PHYSICAL_VALUE)
        assert parameter_type3 is parameter_type1.element_type
        parameter_type3 = self.plib.get_parameter_type(
            ParameterType.ATOMIC, ParameterType.MATERIAL)
        assert parameter_type3 is parameter_type1.key_type

        parameter_type1.check_value(None)
        parameter_type1.check_value(
            {gold:PhysicalValue(10),
             silicone:PhysicalValue(RangeValue(10, 20))})

        raises(ParamTypeError,
            parameter_type1.check_value, 10)
        raises(ParamTypeError,
            parameter_type1.check_value, [10])
        raises(ParamTypeError,
            parameter_type1.check_value, RangeValue(10, 20))
        raises(ParamTypeError,
            parameter_type1.check_value, 'hello')
        raises(ParamTypeError,
            parameter_type1.check_value, PhysicalValue(10))
        raises(ParamTypeError,
            parameter_type1.check_value, [PhysicalValue(10)])
        raises(ParamTypeError,
            parameter_type1.check_value, [gold])
        raises(AssertionError,
            self.plib.get_parameter_type,
            ParameterType.TABLE, ParameterType.INT)

if __name__ == "__main__":
    ParameterTypeTest()



#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/bin/expire_session.py $
$Id: expire_session.py 27873 2006-01-26 17:45:06Z dbinger $

Delete sessions whose access time is older than the age passed in as an
argument (in hours).

This script is meant to be run by cron.
"""

import sys, os
from datetime import datetime, timedelta
from dulcinea.common import get_publisher, get_connection

def main (prog, args):
    usage = "usage: %s [site] [expire time]" % prog

    if len(args) != 2:
        sys.exit(usage)

    site = os.environ['SITE'] = args[0]
    expiration_hours = float(args[1])

    sessions = get_publisher().get_sessions()
    cutoff = datetime.now() - timedelta(hours=expiration_hours)
    for key, session in sessions.items():
        if session.get_access_time() < cutoff:
            del sessions[key]
    get_connection().commit()

if __name__ == '__main__':
    main(sys.argv[0], sys.argv[1:])



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/item.py $
$Id: item.py 27812 2006-01-13 16:52:28Z dbinger $
"""
from dulcinea.base import DulcineaPersistent
from dulcinea.timestamped import Timestamped, reverse_timestamp_sorted
from dulcinea.util import datetime_to_int
from durus.persistent_dict import PersistentDict
from qp.lib.spec import add_getters_and_setters, string
from qp.lib.spec import boolean, mapping, init, require, specify


class Item(DulcineaPersistent, Timestamped):

    title_is = string
    key_is = string
    approved_is = boolean

    def __init__(self):
        assert self.__class__ is not Item, "abstract class"
        Timestamped.__init__(self)
        init(self, approved=False)

    def __cmp__(self, other):
        require(other, Item)
        return cmp(self.get_timestamp(), other.get_timestamp())

    def absorb(self, other):
        assert self.__class__ is other.__class__
        self._p_note_change()
        for attr_name in self.__dict__:
            if attr_name != 'key':
                self.__dict__[attr_name] = other.__dict__[attr_name]

    def set_key(self, key):
        assert self.key is None, "Item already has a key"
        specify(self, key=key)

    def is_approved(self):
        return self.approved

    def can_modify(self, user):
        return user.is_admin()

add_getters_and_setters(Item)


class ItemFolder(DulcineaPersistent):
    """
    Stores a collection of Items and defines an API for storage
    and creation-time-based retrieval.
    """
    items_is = mapping({string:Item}, PersistentDict)

    def __init__(self):
        specify(self, items=PersistentDict())

    def clear(self):
        self.items.clear()

    def add_item(self, item):
        """(item : Item)
        """
        require(item, Item)
        i = 0
        while 1:
            key =  '%s-%s' % (datetime_to_int(item.get_timestamp()), i)
            if not self.items.has_key(key):
                break
            i += 1
        item.set_key(key)
        self.items[key] = item

    def delete_item(self, item):
        """(item : Item)
        """
        require(item, Item)
        if item.get_key() is None:
            raise KeyError, "Item %r not in any database" % item
        if not self.items.has_key( item.key ):
            raise KeyError, "Item %r not in this database" % item
        del self.items[item.get_key()]

    def get_item(self, key):
        """(key : string) -> Item
        """
        return self.items[key]

    def get_items(self):
        """() -> [Item]
        """
        return self.items.values()

    def get_recent_items(self, count=None, include_unapproved=False):
        """(count : int = None, include_unapproved : bool = False) -> [Item]

        Return the 'count' most recent items.  If 'include_unapproved' is True,
        both un-approved and approved items are returned.
        The list returned is in reverse-chronological order by timestamp.
        """
        result = []
        for item in self.items.values():
            if item.is_approved() or include_unapproved:
                result.append(item)
        result = reverse_timestamp_sorted(result)
        if count is not None:
            result = result[:count]
        return result



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/param/errors.py $
$Id: errors.py 27392 2005-09-13 14:28:31Z dbinger $
"""

class ParamError (Exception):
    """Base class for all parameter-related exceptions.
    """

class ParamTypeError (ParamError):
    """Raised on violation of the parameter type system, eg. if someone
    passes an integer for a string parameter, or a number for
    a physical_value parameter.
    """

class ConstraintError (ParamError):
    """Raised when attempting to set a parameter value in violation
    of the parameter's constraint.
    """





"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/param/parameter.py $
$Id: parameter.py 28522 2006-07-24 20:35:21Z dbinger $
"""
from dulcinea.base import DulcineaPersistent
from dulcinea.param.parameter_template import ParameterTemplate
from qp.lib.spec import anything, spec, specify, add_getters_and_setters
from dulcinea.tolerance import Tolerance
from dulcinea.common import htmltext, stringify


class Parameter (DulcineaPersistent):
    """
    Parameters are used to describe properies of things like equipment,
    processes, and process steps within the system.  Each parameter has a
    name, a value and a type.  The name and type are stored in a parameter
    template because they are common between many instances of parameters.
    """
    template_is = spec(
        ParameterTemplate,
        "defines attributes of this parameter such as name, type and "
        "the constraints on the value")
    value_is = spec(
        anything,
        "the value this parameter has (may be None which means no value)")
    tolerance_is = spec(
        (None, Tolerance),
        "the relative tolerance of this parameter (in percent).  May be "
        "None.")

    def __init__(self, template, value=None):
        specify(self, template=template, tolerance=None)
        self.set_value(value)

    def __str__(self):
        if self.value is None:
            return "%s (no value)" % self.get_name()
        elif not self.value:            # empty list/string/etc.
            return "%s (empty value)" % self.get_name()
        else:
            return "%s = %s" % (self.get_name(), self.format())

    def is_required(self):
        return self.template.is_required()

    def is_hidden(self):
        return self.template.is_hidden()

    def is_master(self):
        return 0

    def is_input(self):
        return 0

    def is_parameter(self):
        return 1

    def set_value(self, value):
        """(any)

        May raise TypeError.
        """
        self.template.check_type(value)
        self.value = value

    def allow_tolerance(self):
        return self.template.get_type().allow_tolerance()

    def check_value(self):
        self.template.check_value(self.value)

    def format(self, html=0, show_tolerance=0):
        """Return the parameter's value (and unit, if any) as a useful,
        standalone string.  Return None if the value is undefined."""
        if self.value is None:
            return None

        param_type = self.template.get_type()
        s = param_type.format_value(self.value, html=html)
        if show_tolerance and self.tolerance is not None:
            s += ' ' + htmltext(self.tolerance.format(html=html))
        return s

    def check_type(self):
        """()

        Ensure that 'self.value' is compatible with the parameter type.
        Raise ParamTypeError if not.
        """
        self.get_type().check_value(self.value)

    def simplify(self):
        """() -> (string|int|float|None)
        Return a simplified version of this parameter's value. Physical
        values are converted to use the default unit from the master
        template and are returned as floats. Materials are returned as
        strings (using their names).  If self is an InputTemplate,
        simplify the default value
        """
        return self.get_type().simplify_value(self.value)

    # -- Delegates to our template -------------------------------------

    def is_discrete(self):
        return self.template.is_discrete()

    def is_atomic(self):
        return self.template.is_atomic()

    def is_list(self):
        return self.template.is_list()

    def is_table(self):
        return self.template.is_table()

    def is_array(self):
        return self.template.is_array()

    def is_aggregate(self):
        return self.template.is_aggregate()

    def is_time(self):
        return self.template.is_time()

    def get_master(self):
        return self.template.get_master()

    def get_name(self):
        return self.template.get_name()

    def get_type(self):
        return self.template.get_type()

    def get_title(self):
        return self.template.get_title()

    def get_description(self):
        return self.template.get_description()

    def get_hint(self, html=0):
        if html:
            texttype = htmltext
        else:
            texttype = stringify
        hint = texttype(self.get_description() or "")
        if not (self.is_discrete() or self.get_constraint() is None):
            if hint:
                hint += ', '
            hint += self.template.explain_constraint(html=html)
        return hint

    def get_constraint(self):
        constraint = self.template.get_constraint()
        if self.get_type().is_material():
            materials = set()
            for material in constraint:
                materials.update(material.get_leaves())
            return list(materials)
        else:
            return constraint

    def get_material_constraint(self):
        assert self.get_type().is_material()
        materials = set()
        for material in self.template.get_constraint():
            materials.add(material)
            materials.update(material.get_descendants())
        return list(materials)

    def get_allowed_units(self):
        return self.template.get_allowed_units()

add_getters_and_setters(Parameter)


def get_example_parameters(process_parameters):
    """() -> [Parameter]

    Return a list of example value parameters mirroring 'process_parameters'
    except that parameters lacking values (e.g. input) are duplicated
    with values created using the default value of the parameter.
    """
    example_parameters = []
    if process_parameters:
        for parameter in process_parameters:
            if parameter.is_input():
                example_parameters.append(parameter.create_value(
                    value=parameter.get_default_value()))
            else:
                example_parameters.append(parameter.copy())
    return example_parameters



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/common.py $
$Id: common.py 28619 2006-08-18 20:46:32Z dbinger $
"""
import os
import os.path
import sys
from cStringIO import StringIO
from sancho.utest import UTest
import dulcinea
from dulcinea.site_util import get_config_value

def is_qp():
    return os.environ.get('QP') == '1'

if not is_qp():
    from dulcinea.site_util import SiteModule, parse_address, is_live
    from dulcinea.site_util import is_email_enabled
    dulcinea.local = SiteModule()
    sys.modules['dulcinea.local'] = dulcinea.local
    from quixote.ptl import install
    _used = install
    def get_site_name():
        site = os.environ.get('SITE', 'test')
        return site
    import quixote.publish
    from qp.lib.util import import_object

    from durus.client_storage import ClientStorage
    from durus.file_storage import FileStorage
    from durus.connection import Connection

    def create_publisher(connection=None, **kwargs):
        """
        This can only be called when there is no global publisher.
        If no connection is given, a client Connection is constructed.
        Outside code should only call this if it needs to pass extra
        keyword arguments to the Publisher constructor.
        """
        assert quixote.publish.get_publisher() is None
        site = get_site_name()
        if site == 'test':
            assert connection is None
            publisher_class_name = 'dulcinea.ui.publisher.DulcineaPublisher'
            import dulcinea.test.local
            connection = dulcinea.test.local.make_test_connection()
        else:
            publisher_class_name = get_config_value(
                'publisher', site=site,
                fallback='dulcinea.ui.publisher.DulcineaPublisher')
            if connection is None:
                durus_address = get_config_value('durus_address', site=site)
                host, port = parse_address(durus_address)
                storage = ClientStorage(port=port, host=host)
                connection = Connection(storage=storage)
        publisher_class = import_object(publisher_class_name)
        return publisher_class(connection=connection, **kwargs)

    def get_publisher():
        """() -> Publisher
        Return the global publisher, if it already exists.
        If it does not already exist, make one using a client
        storage if possible.
        """
        return quixote.publish.get_publisher() or create_publisher()

    def open_connection(filename=None, readonly=False):
        """(filename=None|str, readonly=True|False) -> Connection
        This can only be called when there is not yet a global publisher.
        This forces a publisher to be made using a FileStorage connection.
        The storage is readonly if the readonly keyword value is True.
         If the filename is given, it is used as the file for the FileStorage.
        If it is not given, the default filename for the site is used.
        """
        assert quixote.publish.get_publisher() is None
        if filename is None:
            site = get_site_name()
            var_directory = get_config_value('var_directory', site=site)
            filename = os.path.join(var_directory, "%s.durus" % site)
        else:
            assert type(filename) is str
        assert readonly in (True, False)
        storage = FileStorage(filename=filename, readonly=readonly)
        publisher = create_publisher(connection=Connection(storage))
        connection = publisher.get_connection()
        assert readonly is False or connection.storage.fp.mode == 'rb'
        return connection

    from quixote import redirect as quixote_redirect
    def redirect(location, permanent=False):
        quixote_redirect(location, permanent=permanent)
        get_publisher().respond_now()
    from quixote import get_field, get_path, get_session
    from quixote import get_request, get_user, get_response

    from quixote.form.css import BASIC_FORM_CSS
    from quixote.form import Form
    from quixote.form.widget import MultipleSelectWidget, ButtonWidget
    from quixote.form.widget import NumberWidget, FloatWidget, IntWidget
    from quixote.form.widget import OptionSelectWidget, CompositeWidget
    from quixote.form.widget import SingleSelectWidget, RadiobuttonsWidget
    from quixote.form.widget import StringWidget, FileWidget, PasswordWidget
    from quixote.form.widget import SubmitWidget, ResetWidget, HiddenWidget
    from quixote.form.widget import TextWidget, CheckboxWidget, SelectWidget
    from quixote.form.widget import WidgetList, WidgetDict, subname, Widget
    from quixote.form.widget import WidgetValueError
    from quixote.html import htmltag, href, url_quote, nl2br
    from quixote.html import stringify, htmltext, htmlescape
    from quixote.html import url_with_query
    from quixote.util import FileStream, StaticFile
    from quixote.util import StaticDirectory as QuixoteStaticDirectory
    from qp.fill.directory import Directory as QPDirectory
    class Directory(QPDirectory):
        """
        This copies the code for two methods here so that the get_path()-type
        functions can be bound to the quixote versions.
        """
        def _q_traverse(self, path):
            """(path: [string]) -> None|Stream|object

            Traverse a path and return the result.  This is what the
            Publisher calls.  The result will be used to set the body of
            the http response.  If the result is None, the Publisher will
            use not_found() to set the 404 response.  If the result is an
            instance of Stream, it will be written out in in chunks.
            Otherwise, the str() of the result of this call will be used
            as the body of the http response.

            Subclasses don't normally need to override this, but may do so.
            This may be useful to provide access control or to implement
            directories that process the entire path themselves.
            """
            component = path[0]
            name = self._q_translate(component)
            if name is None:
                obj = self._q_lookup(component)
            else:
                obj = getattr(self, name)
            if obj is None:
                not_found('%r?' % component)
            if path[1:]:
                if hasattr(obj, '_q_traverse'):
                    result = obj._q_traverse(path[1:])
                else:
                    not_found()
            elif callable(obj):
                result = obj()
            else:
                result = obj
            if result is None:
                not_found()
            else:
                return result

        def __call__(self):
            """() -> None|Stream|object
            If the _q_traverse() of a Directory reaches this instance,
            and there are no more components, it calls this instance.
            We could return None to cause a 404 response, but we'll try
            adding a slash to the end of the url if it looks like that
            appears to be the intent of the current request.
            """
            if (self._q_translate('') is not None and
                not get_request().get_fields()):
                # Fix missing trailing slash.
                return redirect(get_request().get_environ("PATH_INFO") + "/", permanent=True)
            else:
                return None

    class StaticDirectory(QuixoteStaticDirectory):

        def get_exports(self):
            if self.list_directory:
                yield ('', '_q_index', self.path, None)
            files = os.listdir(self.path)
            files.sort()
            for name in files:
                file_path = os.path.join(self.path, name)
                if os.path.islink(file_path):
                    continue
                if os.path.isdir(file_path):
                    yield (name, None, name + '/', None)
                else:
                    yield (name, None, name, None)

    def set_test_environment(input_string='', **env):
        """
        Make a publisher if necessary, and in any case make sure that
        the conection is readonly.
        Make a request with the given body and environment and
        set the request on the publisher, as if the publisher had
        just started processing the request.
        """
        env.setdefault('SCRIPT_NAME', '')
        env.setdefault('PATH_INFO', '/')
        from quixote.http_request import HTTPRequest
        from quixote.logger import DefaultLogger
        import sys
        def __init__(self, **kwargs):
            self.error_log = sys.stdout
            self.access_log = None
            self.error_email = None
        DefaultLogger.__init__ = __init__
        if quixote.publish.get_publisher() is None:
            if get_site_name() in (None, 'test'):
                get_connection()
            else:
                open_connection(readonly=True)
        publisher = get_publisher()
        storage_fp = getattr(get_connection().get_storage(), 'fp', None)
        assert storage_fp is None or storage_fp.mode == 'rb'

        session_cookie_name = get_publisher().get_session_cookie_name()
        session_key = "test"
        cookies='%s="%s"' % (session_cookie_name, session_key)
        env['HTTP_COOKIE'] = cookies

        publisher._set_request(HTTPRequest(StringIO(input_string), env))

        session = get_publisher().session_manager._create_session()
        session.id = session_key
        get_request().session = session
        get_publisher().get_sessions()[session_key] = session
        def nothing(*args):
            pass
        publisher.abort = nothing
        publisher.commit = nothing
        publisher._clear_request = nothing
        quixote.publish._publisher = publisher

    class CommonTest (UTest):
        """
        Run the tests with a readonly connection for the site.
        """
        def _pre(self):
            UTest._pre(self)
            site = get_site_name()
            assert site in (None, 'test') or not is_live(site)
            assert not is_email_enabled()
            set_test_environment()
            get_connection().abort()

    from dulcinea.sendmail import sendmail


    def get_javascript_code():
        return get_response().javascript_code

    def add_javascript_code(name, javascript):
        get_response().add_javascript(name, javascript)
    from dulcinea.ui.directory import get_crumb_tree
else:
    from qp.pub.common import get_publisher as qp_get_publisher
    def get_publisher():
        publisher = qp_get_publisher()
        if publisher:
            return publisher
        from qp.lib.site import Site
        site = Site(os.environ.get('SITE'))
        publisher = site.get_publisher()
        return publisher
    import dulcinea.uiqp
    import sys
    sys.modules['dulcinea.ui'] = dulcinea.uiqp
    dulcinea.ui = dulcinea.uiqp
    def get_site_name():
        return get_publisher().get_site().get_name()
    from qpy import stringify, h8
    htmltext = h8
    htmlescape = h8.quote
    from qp.fill.html import htmltag, href, nl2br, url_quote, url_with_query
    from qp.pub.common import get_request, get_user, get_response
    from qp.pub.common import redirect, get_path, get_session
    def get_field(x, default=None):
        return get_request().get_field(x, default)
    from qp.fill.css import BASIC_FORM_CSS
    from qp.fill.directory import Directory, get_crumb_tree
    from qp.fill.form import Form, add_javascript_code, get_javascript_code
    from qp.fill.static import FileStream, StaticFile, StaticDirectory
    from qp.fill.widget import MultipleSelectWidget, ButtonWidget
    from qp.fill.widget import NumberWidget, FloatWidget, IntWidget
    from qp.fill.widget import OptionSelectWidget, CompositeWidget
    from qp.fill.widget import SingleSelectWidget, RadiobuttonsWidget
    from qp.fill.widget import StringWidget, FileWidget, PasswordWidget
    from qp.fill.widget import SubmitWidget, ResetWidget, HiddenWidget
    from qp.fill.widget import TextWidget, CheckboxWidget, SelectWidget
    from qp.fill.widget import WidgetList, WidgetDict, subname, Widget
    from qp.fill.widget import WidgetValueError
    from qp.pub.common import header, page

    def set_test_environment(input_string='', **env):
        env.setdefault('SCRIPT_NAME', '')
        env.setdefault('PATH_INFO', '/')
        env.setdefault('SERVER_NAME', 'example.org')
        from qp.pub.hit import Hit
        session_key = "test"
        cookies='%s="%s"' % (get_site_name(), session_key)
        env['HTTP_COOKIE'] = cookies
        hit = Hit(StringIO(input_string), env)
        publisher = get_publisher()
        publisher.set_hit(hit)
        session = publisher.create_session()
        publisher.get_sessions()[session_key] = session
        hit.set_session(session)
        def nothing(*args):
            pass
        publisher.commit = nothing

    class CommonTest (UTest):

        def _pre(self):
            set_test_environment()
            get_connection().abort()

    from qp.mail.send import sendmail
    import dulcinea
    class Local (object):
        def __getattr__(self, name):
            return getattr(get_publisher(), name)

    dulcinea.local = Local()
    sys.modules['dulcinea.local'] = dulcinea.local


def get_fields():
    return get_request().get_fields()

def header(title, *args, **kwargs):
    return get_publisher().header(title, *args, **kwargs)

def page(title, *content, **kwargs):
    return get_publisher().page(title, *content, **kwargs)

def footer(*args, **kwargs):
    return get_publisher().footer(*args, **kwargs)

def not_found(*args, **kwargs):
    return get_publisher().not_found(*args, **kwargs)

def access_denied(body=None):
    return get_publisher().respond('Access Denied',
        body or 'That is not currently allowed.', status=403)

def invalid_query(body='Unable to answer that query.'):
    return get_publisher().respond('Invalid Query',
        body or 'Unable to answer that query.', status=400)

def format_user(user, **kwargs):
    return get_publisher().format_user(user, **kwargs)

def format_user_profile(user):
    return get_publisher().format_user_profile(user)

def format_date(date):
    return get_publisher().format_date(date)

def format_date_time(date):
    return get_publisher().format_date_time(date)

def format_login_agreement():
    return get_publisher().format_login_agreement()

def get_feedback_domain():
    return get_publisher().get_feedback_domain()

def get_feedback_addresses():
    return get_publisher().get_feedback_addresses()

def get_site_title():
    return get_publisher().get_site_title()

def get_connection():
    return get_publisher().get_connection()


_used = [get_publisher, stringify, htmltext, htmlescape, htmltag, href,
         url_quote, nl2br, url_with_query, get_request, get_user, get_response,
         get_field, redirect, get_path, get_session,
         Form, StringWidget, FileWidget, PasswordWidget, TextWidget,
         CheckboxWidget, SelectWidget, SingleSelectWidget, RadiobuttonsWidget,
         MultipleSelectWidget, ButtonWidget, SubmitWidget, ResetWidget,
         HiddenWidget, NumberWidget, FloatWidget, IntWidget,
         OptionSelectWidget, CompositeWidget, WidgetList, WidgetDict,
         subname, Widget, WidgetValueError, FileStream, Directory,
         header, footer, page, StaticFile, StaticDirectory,
         format_user, format_user_profile, format_date, format_date_time,
         format_login_agreement, get_feedback_domain, get_feedback_addresses,
         BASIC_FORM_CSS, CommonTest, sendmail, get_crumb_tree,
         add_javascript_code, get_javascript_code]





#!/www/python/bin/python

"""$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/bin/start-durus.py $
$Id: start-durus.py 26324 2005-03-09 21:32:54Z dbinger $

Script invoked by the site command to start and stop Durus storage server.
"""

import sys, os
from dulcinea.site_util import get_pid_file_name, parse_address, \
     get_config_value, ensure_uid_gid_not_root, is_local
from durus import run_durus, storage_server
ensure_uid_gid_not_root()
site = sys.argv[1]
action = sys.argv[2]

pid_file_name = get_pid_file_name('durus', site)
ip, port = parse_address(get_config_value('durus_address', site=site))
assert is_local(get_config_value('durus_address', site=site))

db = os.path.join(
    get_config_value('var_directory', site=site), '%s.durus' % site)

site_log_dir = os.path.join(get_config_value('log_directory', site=site),
                            site)
durus_log = os.path.join(site_log_dir, 'durus.log')

durus = get_config_value('durus', fallback=run_durus.__file__, site=site)
logginglevel = get_config_value('logginglevel', fallback=20, site=site)

if durus.endswith('c'):
    durus = durus[:-1]

args =  [sys.executable,
         durus,
         '--host', ip,
         '--port', str(port),
         '--file', db,
         '--logginglevel', str(logginglevel),
         '--logfile', durus_log]
if action == 'stop':
    args.append('--stop')
elif action != 'start':
    raise SystemExit, 'action must be "start" or "stop"'

pid = os.fork()
if pid == 0:
    os.execve(args[0], args, {})
else:
    if action == 'start':
        pid_file = open(pid_file_name, 'w')
        pid_file.write("%s" % pid)
        pid_file.close()
        storage_server.wait_for_server(ip, port)



"""$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/site_util.py $
$Id: site_util.py 28106 2006-03-30 16:27:08Z dbinger $

Provides access to site configuration data.
"""

import sys, os, re, socket, errno
try:
    import site_config
except ImportError:
    if 'SITE' in os.environ:
        raise
    # No SITE is set, so we won't really be using a site_config.
    # Put 'site_config' in sys modules so that other site_config
    # imports don't raise ImportError.
    sys.modules['site_config'] = None

def get_config_value(name, fallback=None, site=None):
    if site is None:
        site = os.environ.get('SITE')
    if site in site_config.config:
        return site_config.config[site].get(name, fallback)
    # qp
    from dulcinea.common import get_publisher
    return get_publisher().configuration.get(name, fallback)

def list_sites():
    return site_config.config.keys()

def get_hostname():
    return socket.gethostbyname_ex(socket.gethostname())[0]

def is_staging(site=None):
    return get_config_value('mode', site=site) == 'staging'

def is_live(site=None):
    return get_config_value('mode', site=site) == 'live'

def is_devel(site=None):
    return not (is_live(site) or is_staging(site))

def any_live_sites():
    """Returns True if any sites are live."""
    for site in list_sites():
        if is_live(site):
            return True
    else:
        return False

def any_staging_sites():
    """Returns True if any sites are staging."""
    for site in list_sites():
        if is_staging(site):
            return True
    else:
        return False

def _get_local_interfaces():
    """Return a list of local IP addresses for this machine."""
    # Currently a Linux specific hack. :-(
    addrs = []
    p = os.popen("/sbin/ifconfig")
    for line in p.readlines():
        m = re.search(r'inet (addr:)?(\d+\.\d+\.\d+\.\d+)', line)
        if m:
            addrs.append(m.group(2))
    return addrs

local_interfaces = _get_local_interfaces()

def parse_address(s):
    ip, port = s.strip().split(':')
    if not ip[0].isdigit():
        # not an IP address, resolve it
        ip = socket.gethostbyname(ip)
    return ip, port and int(port)

def is_local(address):
    ip, port = parse_address(address)
    return ip in local_interfaces

def get_root_exports(site=None):
    return get_config_value('root_exports', fallback=[], site=site)

def get_scgi_off(site=None):
    return get_config_value('scgi_off', fallback=[], site=site)

def get_cache_size(site=None, default=10000):
    return int(get_config_value('cache_size', fallback=default, site=site))

_email_pat = re.compile(r'^(.*)\s+<(.*)>$')

def get_administrator_address(site=None):
    administrator = get_config_value('administrator', fallback='', site=site)
    m = _email_pat.match(administrator)
    if not m:
        raise ValueError, '"administrator" in site.conf missing real name'
    realname = m.group(1)
    email = m.group(2)
    return (email, realname)

_email_enabled = False

def enable_email():
    global _email_enabled
    _email_enabled = True

def is_email_enabled():
    """Return true if email should be sent.  This is false for various sites
    and also when tests are being run (for example).
    """
    return _email_enabled

def get_base_path(site=None):
    return get_config_value('base_path', fallback='', site=site)

def get_pid_file_name(daemon, site=None):
    name = '%s.pid' % daemon
    if site:
        name = site + '-' + name
    pid_dir = get_config_value('var_directory', site=site)
    return os.path.join(pid_dir, name)

def write_pid_file(daemon, site=None):
    file = open(get_pid_file_name(daemon, site), 'w')
    file.write(str(os.getpid()))
    file.close()

def change_uid_gid(uid, gid=None):
    "Try to change UID and GID to the provided values"
    # This will only work if this script is run by root.
    # Try to convert uid and gid to integers, in case they're numeric
    import pwd, grp
    try:
        uid = int(uid)
        default_grp = pwd.getpwuid(uid)[3]
    except ValueError:
        uid, default_grp = pwd.getpwnam(uid)[2:4]

    if gid is None:
        gid = default_grp
    else:
        try:
            gid = int(gid)
        except ValueError:
            gid = grp.getgrnam(gid)[2]
    os.setgid(gid)
    os.setuid(uid)

def ensure_uid_gid_not_root():
    if os.getuid() == 0:
        uid = get_config_value('daemon_uid')
        change_uid_gid(uid)
    assert os.getuid() != 0

def log_to(log_file_name):
    log = open(log_file_name, "a", 1)
    os.dup2(log.fileno(), 1)
    os.dup2(log.fileno(), 2)
    os.close(0)

def fetch_database(site, destination=None, source_file=None):
    """
    Pull a copy of the live database for site to the local
    machine using scp.
    If destination is given, it is the name of local file.
    If source file is given, it is a file name
    in the var_directory of the remote host.
    If neither of the keyword arguments are provided, the
    database is pulled from the standard place on the remote machine
    and put in the standard place on the local machine.
    The standard place is <site>.durus in the var_directory.
    """
    remote = get_config_value('live_db', site=site)
    if not destination:
        var_directory = get_config_value('var_directory', site=site)
        destination = os.path.join(var_directory, "%s.durus" % site)
    if remote:
        cmd = 'scp -C -p %s %s' % (remote, destination)
        print cmd
        os.system(cmd)
    return destination

def import_class(name):
    i = name.rfind('.')
    module_name = name[:i]
    class_name = name[i+1:]
    __import__(module_name)
    return getattr(sys.modules[module_name], class_name)

def get_root_directory(site=None):
    root_directory_name = get_config_value('root_directory', site=site)
    return import_class(root_directory_name)()

def get_docroot(site=None):
    if site is None:
        site = os.environ.get('SITE', None)
    sites_directory = get_config_value('sites_directory', site=site,
                                       fallback='')
    return '%s/%s/docroot' % (sites_directory, site)

class SiteModule:
    """
    A module-like object that provides access to site-specific variants
    of a module, delaying the import of the site-specific variants until
    they are actually needed.

    This is used for dulcinea.local.

    Accessing attributes of an instance of this class requires that the
    SITE environment variable is set to the name of a package that
    contains the site-specific module.
    """
    def __init__(self):
        self._loaded = False

    def __getattr__(self, name):
        if not self._loaded and not name.startswith('_'):
            site = os.environ.get('SITE', 'dulcinea.test')
            module_name = site + '.local'
            try:
                __import__(module_name)
                for n, v in vars(sys.modules[module_name]).items():
                    if not n.startswith('_'):
                        setattr(self, n, v)
            except ImportError:
                pass # no local module defined for this site.
            self._loaded = True
        try:
            return self.__dict__[name]
        except KeyError:
            raise AttributeError, name

def is_running(pid):
    try:
        os.kill(int(pid), 0)
        return True
    except OSError, e:
        if e.errno == errno.EPERM:
            return True
    return False



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/local.py $
$Id: local.py 27988 2006-02-17 22:29:13Z rmasse $
"""
from dulcinea.param.parameter_type import ParameterDatabase
from dulcinea.physical_unit import UnitCollection
from dulcinea.inquiry import InquiryDatabase
from dulcinea.link_triple import LinkTripleDatabase
from dulcinea.material import MaterialDatabase, Material
from dulcinea.persistent_session import DulcineaSessionManager
from dulcinea.user import DulcineaUserDatabase
from durus.storage import MemoryStorage
from durus.connection import Connection


def make_test_connection():
    connection = Connection(MemoryStorage())
    connection.get_root().update(dict(
        parameter_db=ParameterDatabase(),
        material_db=MaterialDatabase(),
        standard_units=UnitCollection(),
        link_triple_db=LinkTripleDatabase(),
        user_db=DulcineaUserDatabase(),
        inquiry_db=InquiryDatabase(),
        session_manager=DulcineaSessionManager()))
    mdb = connection.get_root()['material_db']
    mdb.add_material(Material('gold'))
    mdb.add_material(Material('aluminum'))
    mdb.add_material(Material('silver'))
    mdb.add_material(Material('boron'))
    mdb.add_material(Material('silicon'))
    mdb.add_material(Material('blah'))
    connection.commit()
    return connection



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/param/__init__.py $
$Id: __init__.py 28119 2006-03-31 21:57:52Z rmasse $
"""
from dulcinea.common import get_publisher

def get_parameter_db():
    return get_publisher().get_connection().get_root().get('parameter_db')



from os.path import join, islink, samefile, isdir, exists
from os import listdir, symlink, makedirs
from qpy.compile import compile_qpy_files
import dulcinea

def build_ui_mirror(package, path_in_package):
    assert path_in_package.startswith('/uiqp')
    package_path = package.__path__[0]
    relative_uiqp = path_in_package
    relative_ui = '/ui' + relative_uiqp[5:]
    for name in listdir(package_path + relative_ui):
        if name.endswith('~'):
            continue
        if name.endswith('.pyc'):
            continue
        if name.endswith('.pyo'):
            continue
        if name.startswith('.'):
            continue
        if path_in_package == '/uiqp' and name.startswith('__init__.py'):
            continue
        src = join(package_path + relative_ui, name)
        if name.endswith('.ptl'):
            dst = join(package_path + relative_uiqp, name[:-4] + '.qpy')
        else:
            dst = join(package_path + relative_uiqp, name)
        if islink(dst):
            assert samefile(dst, src), (src, dst)
        elif isdir(src):
            if not exists(dst):
                makedirs(dst)
        else:
            try:
                symlink(src, dst)
            except OSError:
                print (src, dst)
                raise
    compile_qpy_files(package_path + relative_uiqp)

build_ui_mirror(dulcinea, '/uiqp')
build_ui_mirror(dulcinea, '/uiqp/lib')
build_ui_mirror(dulcinea, '/uiqp/user')
build_ui_mirror(dulcinea, '/uiqp/form2')
build_ui_mirror(dulcinea, '/uiqp/test')




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_user.py $
$Id: utest_user.py 28600 2006-08-17 14:13:32Z rmasse $
"""
from dulcinea.user import get_realm, DulcineaUser as User
from sancho.utest import raises
from dulcinea.common import CommonTest

class DulcineaUserTest (CommonTest):

    def check_init (self):
        raises(TypeError, User, '!@#$%^')
        User('amk')
        u = User(None)
        u.set_id('foo')
        raises(AssertionError, User('foo').set_id, 'foo')
        raises(TypeError, User(None).set_id, '!foo')
        u = User(None)
        assert u.id == None
        u.set_id('foo')
        assert u.id == "foo"
        raises(AssertionError, u.set_id, 'bar')

    def check_password (self):
        user = User('foo')
        user.set_password("biteme")
        assert user.has_password("biteme")
        assert not user.has_password("dontbiteme")
        assert not user.has_password("biteme", realm='bogus')
        assert user.has_password("biteme", realm=get_realm())
        password1 = user.generate_password()
        assert user.has_password(password1)
        password2 = user.generate_password()
        assert user.has_password(password2)
        assert password1 != password2


if __name__ == "__main__":
    DulcineaUserTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_survey.py $
$Id: utest_survey.py 27784 2006-01-06 19:03:00Z dbinger $
"""
from dulcinea.survey import Question, TextQuestion, SelectQuestion, Surveyable
from dulcinea.survey import QuestionDatabase, TextAnswer, SelectAnswer, Survey
from dulcinea.survey import SurveysAndQuestions, SurveysAndQuestionsDatabase
from dulcinea.user import DulcineaUser
from sancho.utest import UTest


class SurveyTest (UTest):

    class TestSurveyable(Surveyable):

        def __init__(self, question_db=None):
            Surveyable.__init__(self)
            self.question_db = question_db

        def get_new_survey(self):
            return Survey(self.question_db.get_answers([1,2,3]))

    class BadQuestion(Question):
        pass

    def _pre(self):
        self.user = DulcineaUser('u')
        self.questions = [
            TextQuestion("text1 title", sub_title="text sub title"),
            SelectQuestion("select1 title", column_headings=['Yes', 'No']),
            SelectQuestion("select2 title",
                           sub_title="check-all-that-apply question",
                           row_headings=["using machining",
                                         "using bulk ",
                                         "using substrate",
                                         "using substrate",
                                         "equipment supplier",
                                         "material supplier",
                                         "service supplier",
                                         "Other (please specify):"],
                           last_row_has_text=True),
            SelectQuestion("select3 title",
                           sub_title="This example uses a numbered radio set",
                           column_headings=map(str, range(1,10)),
                           row_headings=["Material type",
                                         "Thickness",
                                         "Thickness uniformity",
                                         "Resistivity",
                                         "Flatness",
                                         "Defect level"]),
            SelectQuestion("select4 title",
                           sub_title="This example uses 'range labels'",
                           range_labels = ["Not Important",
                                           "Very Important"],
                           column_headings=map(str, range(1,10)),
                           row_headings=["Material type",
                                         "Thickness",
                                         "Thickness uniformity",
                                         "Resistivity",
                                         "Flatness",
                                         "Defect level"])]
        self.question_db = QuestionDatabase()
        for question in self.questions:
            self.question_db.add(question)

    def check_bad_question(self):
        b1 = self.BadQuestion('bad')
        try:
            b1.get_new_answer()
            assert 0
        except NotImplementedError: pass

    def check_text_question(self):
         t1 = self.question_db.get(1)
         assert t1.get_new_answer().__class__ == TextAnswer
         assert t1.get_title() == "text1 title"
         assert t1.get_sub_title() == 'text sub title'

    def check_select_question(self):
         s1 = self.question_db.get(2)
         assert s1.get_new_answer().__class__ == SelectAnswer
         assert s1.get_title() == "select1 title"
         assert s1.get_sub_title() == None
         assert s1.get_column_headings() == ['Yes', 'No']
         assert s1.get_row_headings() == []
         assert s1.get_range_labels() == []
         s2 = self.question_db.get(3)
         assert s2.get_sub_title() == "check-all-that-apply question"
         assert s2.get_row_headings() == [
             "using machining",
             "using bulk ",
             "using substrate",
             "using substrate",
             "equipment supplier",
             "material supplier",
             "service supplier",
             "Other (please specify):"]
         s4 = self.question_db.get(5)
         assert s4.get_range_labels() == ["Not Important", "Very Important"]

    def check_text_answer(self):
         self.t1 = self.question_db.get(1)
         assert self.t1.get_new_answer().__class__ == TextAnswer
         a1 = self.t1.get_new_answer()
         assert a1.get_question() == self.t1
         assert a1.get_answer() == None
         assert a1.has_answer() == False
         a1.set_answer('the answer')
         assert a1.get_answer() == 'the answer'
         assert a1.has_answer() == True

    def check_select_answer(self):
         self.s1 = self.question_db.get(2)
         assert self.s1.get_new_answer().__class__ == SelectAnswer
         a2 = self.s1.get_new_answer()
         assert a2.get_question() == self.s1
         assert a2.get_answer() == None
         assert a2.has_answer() == False
         a2.set_answer(['1','2','3'])
         assert a2.get_answer() == ['1','2','3']
         assert a2.has_answer() == True
         assert a2.get_other_text() == None
         a2.set_other_text('other')
         assert a2.get_other_text() == 'other'

    def check_survey(self):
        from qp.lib.spec import get_spec_problems
        answers = self.question_db.get_answers([1,2,3,4,5])
        for j, answer in enumerate(answers):
            problems = get_spec_problems(answer)
            if problems:
                print 'answer', j
                for problem in problems:
                    print problem

        s1 = Survey(answers)
        str(s1)
        assert len(s1) == 5
        assert len(s1.get_questions()) == 5
        assert len(s1.get_answers()) == 5
        timestamp = s1.get_timestamp()
        s1.set_timestamp()
        assert s1.get_timestamp() != timestamp
        assert s1.get_user() == None
        s1.set_user(self.user)
        assert s1.get_user().get_id() == 'u'
        assert not s1.has_answers()
        a1 = s1.get_answers()[0]
        a1.set_answer('the answer')
        assert s1.has_answers()
        s1.set_answers(s1.get_answers(), self.user)

    def check_surveyable(self):
        sv = self.TestSurveyable(question_db=self.question_db)
        assert sv.get_surveys() == []
        sv.add_survey(Survey(self.question_db.get_answers([1,2,3])))
        sv.get_survey(0)
        sv.get_new_survey()
        sv2 = Surveyable()
        try:
            sv2.get_new_survey()
            assert 0
        except NotImplementedError: pass

    def check_surveys_and_questions(self):
        questions = [TextQuestion(
            "Do you wish to be included?",
            sub_title=("If you wish to be entered, please "
                       "enter your contact information below so you can be "
                       "notified.")),
                     SelectQuestion(
            "Are there factors more critical "
            "than hypothetical issues?", column_headings=['Yes', 'No']),
                     SelectQuestion(
            "Please choose one or more of the following categories to "
            "represent your club's activities.",
            row_headings=["maker using dirt",
                          "maker using glass",
                          "maker using metal",
                          "maker using wood",
                          "equipment supplier",
                          "material supplier",
                          "service supplier",
                          "Other (please specify):"],
            sub_title="This is a 'check-all-that-apply' style question",
            last_row_has_text=True),
                     SelectQuestion(
            "What are the key  issues?",
            sub_title="This example uses a numbered radio set",
            column_headings=map(str, range(1,10)),
            row_headings=["type", "ness", "uniformity",
                          "blah", "Flatness", "level"]),
                     SelectQuestion(
            "What are the key?",
            sub_title="This example uses 'range labels' in addition to numbers",
            range_labels = ["Not Important", "Very Important"],
            column_headings=map(str, range(1,10)),
            row_headings=["type", "ness", "uniformity",
                          "blah", "Flatness", "level"])]

        sq = SurveysAndQuestions('test', questions)
        assert sq.get_id() == 'test'
        assert len(sq.get_new_survey().get_questions()) == len(questions)
        sq_db = SurveysAndQuestionsDatabase()
        sq_db.add(sq)
        assert sq_db.get('test') is sq
        sq_db.get_mapping()['test'] is sq

if __name__ == "__main__":
    SurveyTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_timestamped.py $
$Id: utest_timestamped.py 27147 2005-08-03 13:59:08Z dbinger $
"""
from sancho.utest import UTest
from dulcinea.timestamped import Timestamped, timestamp_sorted
from dulcinea.timestamped import reverse_timestamp_sorted


class Test (UTest):

    def a(self):
        x = Timestamped()
        y = Timestamped()
        assert timestamp_sorted([x, y]) == [x,y]
        assert reverse_timestamp_sorted([x, y]) == [y,x]
        x.set_timestamp()
        assert timestamp_sorted([x, y]) == [y,x]

if __name__ == "__main__":
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/param/test/utest_parameter_database.py $
$Id: utest_parameter_database.py 28105 2006-03-30 16:24:19Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.param import get_parameter_db
from dulcinea.param.parameter_template import MasterTemplate
from dulcinea.param.parameter_type import get_string_parameter_type
from sancho.utest import raises

class ParameterDatabaseTest (CommonTest):

    def check_master_templates (self):
        if get_parameter_db() is None:
            return
        parameter_db = get_parameter_db()
        raises(TypeError, parameter_db.add_master_template, 'foo')

        s = get_string_parameter_type()
        self.foo = MasterTemplate('foo',s)
        self.bar = MasterTemplate('bar',s)
        raises(ValueError, parameter_db.rename_master_template, self.bar,'foo')

        parameter_db.add_master_template(self.foo)
        parameter_db.add_master_template(self.bar)
        raises(ValueError, parameter_db.add_master_template, self.foo)

        raises(ValueError, parameter_db.rename_master_template, self.bar, 'foo')
        parameter_db.rename_master_template(self.bar,'baz')
        assert self.bar == parameter_db.get_master_template('baz')


if __name__ == "__main__":
    ParameterDatabaseTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/param/test/utest_list.py $
$Id: utest_list.py 28105 2006-03-30 16:24:19Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.param import get_parameter_db
from dulcinea.param.parameter import Parameter
from dulcinea.param.parameter_list import ParameterList
from dulcinea.param.parameter_template import MasterTemplate
from dulcinea.param.parameter_type import ParameterType
from dulcinea.physical_value import PhysicalValue
from sancho.utest import raises

class TestParameterList (CommonTest):

    def _pre(self):
        CommonTest._pre(self)
        if get_parameter_db() is None:
            return
        get_type = get_parameter_db().get_parameter_type
        self.int_type = get_type(ParameterType.ATOMIC,
                                 ParameterType.INT)
        sec = PhysicalValue(1, 's').unit
        self.pv_type = get_type(ParameterType.ATOMIC,
                                ParameterType.PHYSICAL_VALUE,
                                element_sub_type=sec)
        um =  PhysicalValue(1, 'um').unit
        self.pv_um_type = get_type(ParameterType.ATOMIC,
                                ParameterType.PHYSICAL_VALUE,
                                element_sub_type=um)
        self.str_list_type = get_type(ParameterType.LIST,
                                      ParameterType.STRING)
        self.str_int_table_type = get_type(ParameterType.TABLE,
                                           ParameterType.INT,
                                           key_type=ParameterType.STRING)


    def check_basic (self):
        if get_parameter_db() is None:
            return
        template = MasterTemplate('downtime', self.pv_type)
        parameter = template.create_value()
        ParameterList([parameter])
        ParameterList([template])
        ParameterList([parameter])
        ParameterList([template])
        raises(TypeError, ParameterList, [parameter, object()])

    def check_filtering (self):
        if get_parameter_db() is None:
            return
        template = MasterTemplate("a", self.int_type)
        parameter = Parameter(MasterTemplate("b", self.pv_type))
        plist = ParameterList([template, parameter])
        assert list(plist.get_value_parameters()) == [parameter]
        assert list(plist.get_templates()) == [template]
        assert list(plist.get_visible()) == [template, parameter]
        assert list(plist + []) == list(plist)

if __name__ == "__main__":
    TestParameterList()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/__init__.py $
$Id: __init__.py 28638 2006-08-22 13:20:02Z dbinger $
"""
__version__ = "0.15"

try:
    import dulcinea.common
except ImportError:
    print "Dulcinea", __version__



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/ui/directory.py $
$Id: directory.py 28617 2006-08-18 18:45:36Z dbinger $
"""
from dulcinea.common import get_request, get_response, get_path, is_qp
from qp.fill.directory import get_directory_path

assert not is_qp()

def get_crumb_tree():
    def get_path_directory_list():
        app_root = get_request().get_environ('SCRIPT_NAME', '')
        path = get_path()
        path_components = path[len(app_root):].split('/')
        namespace_stack = get_directory_path()
        if get_response().get_status_code() == 200:
            offset = -1
        else:
            offset = 0
        return [('./' + '../' * (len(namespace_stack) - k + offset),
                 namespace_stack[k])
                for k in range(min(len(path_components), len(namespace_stack)))]

    def gen_crumb_tree_menu_items(path, directory):
        for component, name, crumb, title in directory.get_exports():
            if crumb: # skip crumbs and menus if no crumb
                css_class = 'menu_' + component.replace('.', '_')
                yield (path + component, crumb, title, css_class)


    """() - > [[(path:str, crumb:str, title:str, css_class:str)]]
    Each list in the result is a menu of exports for a directory on the
    current path with a derived css class name for each menu entry.
    """
    crumb_tree = []
    for path, directory in get_path_directory_list():
        menu = list(gen_crumb_tree_menu_items(path, directory))
        if menu:
            crumb_tree.append(menu)
    return crumb_tree
